
#include "x.h"
#include "xm.h"
#include "xpex.h"
#include "xmbX.h"
#include <math.h>
#include "book_utils.h"
#include "defs.h"
#include "off.h"
#include "externs.h"

#define DEGRAD( _a )  ( (double)( _a ) * 3.1415926 / 180.0 )

typedef struct _Vertex {
    float x, y, z;
} Vertex;

typedef Vertex Vector;

typedef struct _Edge {
    int e;
} Edge;

typedef struct _Facet {
    int count;
    Edge *edges;
} Facet;

extern XtAppContext app_context;

XtActionProc CanvasButtonPress(), CanvasButtonRelease(),
             CanvasB1Motion(), CanvasB2Motion(), CanvasB3Motion(),
             CanvasMessage(), CanvasExpose(), CanvasConfigure(), CanvasResize(),
             CanvasUp(), CanvasDown(), CanvasLeft(), CanvasRight(),
             CanvasReset(),
             CanvasWireframe(), CanvasCull(), CanvasEdge(), CanvasShadow(),
             CanvasColor(), CanvasLight(),
             CanvasMaterial(), CanvasMaterialPredef(),
             CanvasTaper(), CanvasTwist(), CanvasBend(),
             CanvasStepBack(), CanvasPlay(), CanvasStepForw(), CanvasStop(),
             CanvasRecord(), CanvasSavePPM(),
             CanvasHardcopy(),
             CanvasMaterialPredefQuit(), CanvasVideoQuit(), CanvasQuit();

XtActionsRec canvas_actions[] =
{
    { "Bdn", (XtActionProc)CanvasButtonPress },
    { "Bup", (XtActionProc)CanvasButtonRelease },
    { "B1mt", (XtActionProc)CanvasB1Motion },
    { "B2mt", (XtActionProc)CanvasB2Motion },
    { "B3mt", (XtActionProc)CanvasB3Motion },
    { "Resize", (XtActionProc)CanvasResize },
    { "Expose", (XtActionProc)CanvasExpose },
/*  { "ClientMessage", (XtActionProc)CanvasMessage }, */
    { "Configure", (XtActionProc)CanvasConfigure },
    { "Up", (XtActionProc)CanvasUp },
    { "Down", (XtActionProc)CanvasDown },
    { "Left", (XtActionProc)CanvasLeft },
    { "Right", (XtActionProc)CanvasRight },
    { "Reset", (XtActionProc)CanvasReset },
    { "Print",(XtActionProc)CanvasHardcopy },
    { "Quit", (XtActionProc)CanvasQuit },
};


String canvas_translations =
   "<BtnDown>:        Bdn() \n\
    <BtnUp>:        Bup() \n\
    <Btn1Motion>:    B1mt() \n\
    <Btn2Motion>:    B2mt() \n\
    <Btn3Motion>:    B3mt() \n\
    <ResReq>:        Resize() \n\
    <Expose>:        Expose() \n\
    <Configure>:    Configure() \n\
    <Key>k:            Up() \n\
    <Key>j:            Down() \n\
    <Key>h:            Left() \n\
    <Key>l:            Right() \n\
    <Key>r:            Reset() \n\
    <Key>p:            Print() \n\
    <Key>q:            Quit()";


extern Display      *display;
Window               window;
XWindowAttributes    xwattrs;
Pixmap               pixmap;
XStandardColormap    cmap_info;
GC                   gc;
Multibuffer          buffers[2];
XVisualInfo          vis_info;

PEXRenderer				renderer;
PEXRendererAttributes	rattrs;
PEXMatrix				init_matrix, local_matrix, shad_local_matrix;

unsigned int nvertices, nfacets;
Facet	*facets   = NULL;
Vertex  *vertices = NULL;
Vertex  *morph_vertices = NULL;
PEXColorRGB *facet_colors;
int      shape_hint, ignore_edges, contour_hint, color_type;
unsigned int facet_attributes, vertex_attributes;

PEXStructure top_struct, empty;
PEXStructure object, object_shadow, axis, plane;

int playing = 0;
int stopping = 1;
int recording = 0;

int draw_shadow = 0;
int draw_axis = 1;
int wireframe = 0;
int culling = 1;
int edges = 0;
extern int verbose;
extern int using_drawable;
Drawable drawable;

PEXCoord        view_ref_pt, view_ref_pt_a;
PEXVector       view_up_vec, view_up_vec_a;
PEXVector       view_plane_normal, view_plane_normal_a;
PEXCoord2D      view_window[2], view_window_a[2];
PEXNPCSubVolume viewport, viewport_a;
PEXViewEntry    view, view_a;
double          view_plane, front_plane, back_plane;
double          view_plane_a, front_plane_a, back_plane_a;
PEXCoord        proj_ref_point, proj_ref_point_a;
int             perspective, perspective_a;

int n_prp_points=0;
PEXVector *prp_points;

int n_vpn_points=0;
PEXVector *vpn_points;

int n_rotation=0, n_translation=0, n_scale=0;
float rot_x, rot_y, rot_z;
PEXVector translation, scale;

extern Widget toplevel, w_canvas;
Widget w_dialog;

PEXColorSpecifier colorSpec;

Boolean colorTypeIndexed, colorTypeRGB,
        colorTypeHSV, colorTypeHLS, colorTypeCIE;

PEXReflectionAttributes refl_props;

PEXColor        colorMaterial, colorEdge, colorShadow;

int             connect_offset = 0;

typedef struct _material_predef {
    PEXColorRGB ambientColor;
    PEXColorRGB diffuseColor;
    PEXColorRGB specularColor;
    float       shininess;
    float       transparency;
    float       shininess_conc;
} material_predef;

material_predef convert[45];

int             numLightsON = 3;
PEXTableIndex   lightsON[4] = { 1, 2, 3 };
int             numLightsOFF = 1;
PEXTableIndex   lightsOFF[4] = { 4 };
int             light_switches[4] = { 1, 1, 1, 0 };
PEXLightEntry   lights[4];

int cur_shading = PEXSurfaceInterpColor;
int cur_refl    = PEXReflectionSpecular;

Atom xa_WM_DELETE_WINDOW;
Atom xa_MOTIF_WM_MESSAGES;
Atom xa_WM_PROTOCOLS;
Atom protocols[2];


void
initPex( Obj )
OFFObjDesc *Obj;
{
    Colormap             colormap;
    PEXColorApproxEntry  capx_info;
    XSetWindowAttributes wattrs;
    XSizeHints           window_hints;

    PEXExtensionInfo     *pexinfo;
    char                 err[80];
    int                  type;
    Position             x, y;
    int                  i, j, k;
    extern int           object_defined;
    PEXColor             color;

    char *namestr;
    extern int no_darea, no_move, no_background;

    /* See if a PEX extension exists and initialize PEXlib. */
    if ( PEXInitialize( display, &pexinfo, 80, err ) != 0 ) {
        fprintf( stderr, "No PEX extension.\n" );
        fprintf( stderr, "%s\n", err );
        exit(1);
    }
    if( verbose )
    fprintf( stderr, "ViewObject: Initialized PEX\n" );

    /* Determine the best visual to use. */
    type = ora_find_best_visual( display, &vis_info );
    /* If type == PEXTrue after this statement */
    /* then pixmap drawable is available.   */
    if( verbose )
    fprintf( stderr, "ViewObject: Determined best Visual\n" );

    if( vis_info.depth != 8 )
        ora_create_colormap( display, &vis_info, &cmap_info, &capx_info );
    else {
        /* Get a standard colormap for the visual. */
        if ( !ora_get_standard_colormap( display, &vis_info, &cmap_info ) ) {
            fprintf ( stderr, "Cannot find a standard colormap\n" );
            exit(1);
        }
        ora_set_stdcmap_approx( &vis_info, &cmap_info, &capx_info );
    }
    if( verbose )
    fprintf( stderr, "ViewObject: Created Colomap\n" );

    if( !object_defined )
        map_dialog(toplevel, &x, &y);
/*
    namestr = (char *) malloc( 18 + strlen( Obj->FirstProp->PropFileName ) );
    strcpy( namestr, "View OFF Object: " );
    strncat( namestr, Obj->FirstProp->PropFileName, strlen(Obj->FirstProp->PropFileName)-5 );
*/
    if( !no_darea ) {
    w_dialog = XtVaCreatePopupShell( "View OFF Object",
        xmDialogShellWidgetClass,
        toplevel,
        XmNautoUnmanage, False,
        XmNdefaultPosition, False,
        XmNx, x,
        XmNy, y+50,
        NULL );

    /* Set the visual, depth, and colormap for the shell. */
    XtVaSetValues( w_dialog,
        XtNvisual, vis_info.visual,
        XtNdepth, vis_info.depth,
        XtNcolormap, cmap_info.colormap,
        XmNx, x,
        XmNy, y+50,
        NULL );

    w_canvas = XtVaCreateManagedWidget( "canvas", xmDrawingAreaWidgetClass,
        w_dialog,
        XmNwidth, 600,
        XmNheight, 600,
        NULL );

    XtManageChild( w_dialog );
    window = XtWindow( w_canvas );

    XtAppAddActions( app_context, canvas_actions, XtNumber( canvas_actions ));
    XtOverrideTranslations( w_canvas,
        XtParseTranslationTable( canvas_translations ));
    }
    else {
        window = ora_create_window( display, DefaultRootWindow( display ),
            &vis_info, &cmap_info, "600x600", "View OFF Object", "OFF" );
        XSelectInput( display, window,
            ExposureMask | KeyPressMask | StructureNotifyMask |
            ButtonPressMask | ButtonMotionMask | ButtonReleaseMask );
    }

    /* Determine the best drawable to use. */
    if( (type == PEXTrue) && (using_drawable == PIXMAP) )
        type = PEXPixmapDrawable;
    else if( using_drawable == PIXMAP ) {
        fprintf( stderr, "No pixmap drawable available.\n" );
        fprintf( stderr, "Switching to a window drawable\n" );
        type = PEXWindowDrawable;
    }
    else if( using_drawable == BUFFER ) {
        if( ora_create_mbx_buffers( display, window, 2, buffers ) )
            type = PEXBufferDrawable;
        else
            type = PEXWindowDrawable;
    }
    else
        type = PEXWindowDrawable;
    if( verbose )
    fprintf( stderr, "ViewObject: Determined best Drawable\n" );

    if( verbose ) {
    switch ( type ) {
    case PEXWindowDrawable:
        fprintf( stderr, "Using:\tWindow Drawable\n" );
        break;
    case PEXPixmapDrawable:
        fprintf( stderr, "Using:\tPixmap Drawable\n" );
        break;
    case PEXBufferDrawable:
        fprintf( stderr, "Using:\tBuffer Drawable\n" );
        break;
    }
    }

#if( 0 ) 
    xa_WM_DELETE_WINDOW  = XInternAtom( display, "WM_DELETE_WINDOW", False );
    xa_MOTIF_WM_MESSAGES = XInternAtom( display, "_MOTIF_WM_MESSAGES", False );
    xa_WM_PROTOCOLS      = XInternAtom( display, "WM_PROTOCOLS", False );
    protocols[0] = xa_WM_DELETE_WINDOW;
    protocols[1] = xa_MOTIF_WM_MESSAGES;
    XSetWMProtocols( display, window, protocols, 2 );
#endif		/*  */
    XGetWindowAttributes( display, window, &xwattrs );
    if( type == PEXBufferDrawable ) {
        using_drawable = BUFFER;
        drawable = buffers[1];
    }
    else if( type == PEXPixmapDrawable ) {
        pixmap = XCreatePixmap( display, window, xwattrs.width, xwattrs.height,
            xwattrs.depth );
        if( pixmap == (Pixmap)NULL ) {
            fprintf( stderr, "Cant allocate a pixmap.\n" );
            exit(1);
        }
        using_drawable = PIXMAP;
        drawable = pixmap;
    }
    else {
        /* Set window attributes. */
        wattrs.backing_store = NotUseful;
        wattrs.background_pixel = 0;
        wattrs.border_pixel = 0;
        wattrs.colormap = cmap_info.colormap;
        XChangeWindowAttributes( display, window,
            CWBackingStore | CWBackPixel | CWBorderPixel | CWColormap,
            &wattrs );
        using_drawable = WINDOW;
        drawable = window;
    }
    gc = DefaultGC( display, DefaultScreen(display) );

    /* Create the PEX renderer. */
    if ( IMMED_MODE_SPT( pexinfo ) ) {
        renderer = ora_setup_renderer( display, window, &capx_info,
            &rattrs );
    } else {
        fprintf( stderr, "Immediate mode not supported.\n" );
        exit(1);
    }
    if( verbose )
    fprintf( stderr, "ViewObject: Created PEX Renderer\n" );

    colorTypeIndexed = ora_enum_type_supported( display, window,
        PEXETColorType, PEXColorTypeIndexed );
    colorTypeRGB = ora_enum_type_supported( display, window,
        PEXETColorType, PEXColorTypeRGB );
    colorTypeHSV = ora_enum_type_supported( display, window,
        PEXETColorType, PEXColorTypeHSV );
    colorTypeHLS = ora_enum_type_supported( display, window,
        PEXETColorType, PEXColorTypeHLS );
    colorTypeCIE = ora_enum_type_supported( display, window,
        PEXETColorType, PEXColorTypeCIE );

    build_plane();
    build_axis();
    build_object_shadow( Obj );
    build_object( Obj );

    top_struct = PEXCreateStructure( display );
    if( verbose )
    fprintf( stderr, "top\tstructure id = %d\n", top_struct );
    empty = PEXCreateStructure( display );
    if( verbose )
    fprintf( stderr, "empty\tstructure id = %d\n", empty );

    /* Initialize view */
    initialize_view();
    PEXSetViewIndex( display, top_struct, PEXOCStore, 1 );

    if( !no_background )
    PEXExecuteStructure( display, top_struct, PEXOCStore, plane );

    PEXSetInteriorStyle( display, top_struct, PEXOCStore,
        PEXInteriorStyleSolid );
    PEXSetFacetDistinguishFlag( display, top_struct, PEXOCStore, True );

    PEXLabel( display, top_struct, PEXOCStore, SHADOW_LABEL );
    PEXNoop( display, top_struct, PEXOCStore );

    initialize_material( Obj );
    initialize_material_predef();
    initialize_light_table( Obj );

    /* Initialize depth cueing table. */
    initialize_depth_cue_table();
    PEXLabel( display, top_struct, PEXOCStore, DEPTH_CUE_LABEL );
    PEXSetDepthCueIndex( display, top_struct, PEXOCStore, 0 );

    /* Set the renderer's Subvolume. */
    rattrs.npc_subvolume.min.x = 0; rattrs.npc_subvolume.max.x = 1;
    rattrs.npc_subvolume.min.y = 0; rattrs.npc_subvolume.max.y = 1;
    rattrs.npc_subvolume.min.z = 0; rattrs.npc_subvolume.max.z = 1;
    /* Set the renderer's Viewport. */
    rattrs.viewport.min.x = 10;     rattrs.viewport.max.x = 590;
    rattrs.viewport.min.y = 10;     rattrs.viewport.max.y = 590;
    rattrs.viewport.min.z =  0;     rattrs.viewport.max.z = 1;
    rattrs.viewport.use_drawable = True;
    PEXChangeRenderer( display, renderer,
        PEXRANPCSubVolume | PEXRAViewport, &rattrs );

    PEXLabel( display, top_struct, PEXOCStore, CULL_LABEL );
    PEXSetFacetCullingMode( display, top_struct, PEXOCStore, PEXBackFaces );

    PEXLabel( display, top_struct, PEXOCStore, WIREFRAME_LABEL );
    PEXSetSurfaceEdgeFlag( display, top_struct, PEXOCStore, PEXFalse );
    SET_COLOR( 0.0, 1.0, 0.0, colorEdge );
    PEXSetSurfaceEdgeColor( display, top_struct, PEXOCStore,
        PEXColorTypeRGB, &colorEdge );

    PEXLabel( display, top_struct, PEXOCStore, SHADING_METHOD_LABEL );
    PEXSetSurfaceInterpMethod( display, top_struct, PEXOCStore, cur_shading );
    PEXSetBFSurfaceInterpMethod( display, top_struct, PEXOCStore, cur_shading );

    PEXLabel( display, top_struct, PEXOCStore, INTERIOR_LABEL );
    PEXSetInteriorStyle( display, top_struct, PEXOCStore,
        PEXInteriorStyleSolid );
    PEXSetBFInteriorStyle( display, top_struct, PEXOCStore,
        PEXInteriorStyleSolid );

    PEXLabel( display, top_struct, PEXOCStore, REFL_EQN_LABEL );
    PEXSetReflectionModel( display, top_struct, PEXOCStore, cur_refl );
    PEXSetBFReflectionModel( display, top_struct, PEXOCStore, cur_refl );

    PEXLabel( display, top_struct, PEXOCStore, OBJECT_LABEL );
    PEXExecuteStructure( display, top_struct, PEXOCStore, object );
}


void
initialize_material( Obj )
OFFObjDesc *Obj;
{
    PEXColor color;

    OFFProperty *props = Obj->FirstProp;

    refl_props.ambient = 0.4;
    refl_props.diffuse = 0.4;
    refl_props.specular = 0.5;
    refl_props.specular_conc = 3.0;
    refl_props.transmission = 0;
    refl_props.specular_color.type = PEXColorTypeRGB;
    SET_COLOR( 1.0, 1.0, 1.0, refl_props.specular_color.value );
    SET_COLOR( 1.0, 0.2, 0.5, colorMaterial );
    SET_COLOR( 0.3, 0.3, 0.3, color );

    while( props != NULL ) {
        char  *ptr;    long  *lptr;    float *fptr;    short *sptr;

        if( strcmp( props->PropName, "ambient_coef" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                refl_props.ambient = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "diffuse_coef" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                refl_props.diffuse = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "specular_coef" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                refl_props.specular = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "specular_power" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                refl_props.specular_conc = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "specular_color" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                refl_props.specular_color.value.rgb.red   = *fptr++;
                refl_props.specular_color.value.rgb.green = *fptr++;
                refl_props.specular_color.value.rgb.blue  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "transmission" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                refl_props.transmission = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "material_colors" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                colorMaterial.rgb.red   = *fptr++;
                colorMaterial.rgb.green = *fptr++;
                colorMaterial.rgb.blue  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        props = props->NextProp;
    }

    PEXLabel( display, top_struct, PEXOCStore, MATERIAL_COLOR_LABEL );
    PEXSetSurfaceColor( display, top_struct, PEXOCStore,
        PEXColorTypeRGB, &colorMaterial );
    PEXSetBFSurfaceColor( display, top_struct, PEXOCStore,
        PEXColorTypeRGB, &color );

    PEXLabel( display, top_struct, PEXOCStore, REFL_PROPS_LABEL );
    PEXSetReflectionAttributes( display, top_struct, PEXOCStore, &refl_props );
    PEXSetBFReflectionAttributes( display, top_struct, PEXOCStore, &refl_props);
}


void
initialize_material_predef()
{
    #include "metal.h"
    #include "plastic.h"
    #include "rubber.h"
    #include "stone.h"
    #include "material.h"

    material_predef material_default = {
        1.0, 0.2, 0.5,
        1, 1, 1,
        1, 1, 1,
        0.5,
        0.0,
        3.0
    };

    convert[ 0] = brass;
    convert[ 1] = bronze;
    convert[ 2] = bronzePolished;
    convert[ 3] = chrome;
    convert[ 4] = copper;
    convert[ 5] = copperPolished;
    convert[ 6] = gold;
    convert[ 7] = goldPolished;
    convert[ 8] = pewter;
    convert[ 9] = silver;
    convert[10] = silverPolished;
    convert[11] = blackPlastic;
    convert[12] = bluePlastic;
    convert[13] = cyanPlastic;
    convert[14] = greenPlastic;
    convert[15] = greyPlastic;
    convert[16] = pinkPlastic;
    convert[17] = redPlastic;
    convert[18] = whitePlastic;
    convert[19] = yellowPlastic;
    convert[20] = blackRubber;
    convert[21] = blueRubber;
    convert[22] = cyanRubber;
    convert[23] = greenRubber;
    convert[24] = magentaRubber;
    convert[25] = orangeRubber;
    convert[26] = redRubber;
    convert[27] = whiteRubber;
    convert[28] = yellowRubber;
    convert[29] = emerald;
    convert[30] = jade;
    convert[31] = obsidian;
    convert[32] = pearl;
    convert[33] = ruby;
    convert[34] = turquoise;
    convert[35] = material_default;
    convert[36] = brown;
    convert[37] = darkbrown;
    convert[38] = gray;
    convert[39] = glass;
    convert[40] = green;
    convert[41] = red;
    convert[42] = sea;
    convert[43] = white;
}


#define NORMALIZE( _v ) \
{ \
    double    mag; \
\
    mag = 1.0 / sqrt((_v).x * (_v).x \
          + (_v).y * (_v).y + (_v).z * (_v).z); \
    (_v).x *= mag, (_v).y *= mag, (_v).z *= mag; \
}


void
initialize_light_table( Obj )
OFFObjDesc *Obj;
{
    int i;
    OFFProperty *props = Obj->FirstProp;

    lights[AMBIENT].type = PEXLightAmbient;
    lights[AMBIENT].color.type = PEXColorTypeRGB;
    SET_COLOR( 1.0, 1.0, 1.0, lights[AMBIENT].color.value );

    lights[DIRECTIONAL].type = PEXLightWCVector;
    lights[DIRECTIONAL].direction.x = -1;
    lights[DIRECTIONAL].direction.y =  0;
    lights[DIRECTIONAL].direction.z =  0;
    NORMALIZE( lights[DIRECTIONAL].direction );
    lights[DIRECTIONAL].color.type = PEXColorTypeRGB;
    SET_COLOR( 1.0, 1.0, 1.0, lights[DIRECTIONAL].color.value );

    lights[POSITIONAL].type = PEXLightWCPoint;
    lights[POSITIONAL].point.x = 1.0;
    lights[POSITIONAL].point.y = 2.0;
    lights[POSITIONAL].point.z = 0.7;
    lights[POSITIONAL].attenuation1 = 0;
    lights[POSITIONAL].attenuation2 = 1;
    lights[POSITIONAL].color.type = PEXColorTypeRGB;
    SET_COLOR( 1.0, 1.0, 1.0, lights[POSITIONAL].color.value );

    /* Define the spot light. */
    lights[SPOT].type = PEXLightWCSpot;
    lights[SPOT].point.x = -4.;
    lights[SPOT].point.y = -4.;
    lights[SPOT].point.z = -3.;
    lights[SPOT].direction.x = -1;
    lights[SPOT].direction.y = -1;
    lights[SPOT].direction.z = -1;
    NORMALIZE( lights[SPOT].direction );
    lights[SPOT].point.x = 0.5;
    lights[SPOT].point.y = 0.5;
    lights[SPOT].point.z = 2.0;
    lights[SPOT].direction.x = 0;
    lights[SPOT].direction.y = 0;
    lights[SPOT].direction.z = 1;
    NORMALIZE( lights[SPOT].direction );
    lights[SPOT].attenuation1 = 0.5;
    lights[SPOT].attenuation2 = 0.1;
    lights[SPOT].concentration = 0;
    lights[SPOT].spread_angle = M_PI_2;
    lights[SPOT].color.type = PEXColorTypeRGB;
    SET_COLOR( 1.0, 1.0, 1.0, lights[SPOT].color.value );

    while( props != NULL ) {
        char  *ptr;    long  *lptr;    float *fptr;    short *sptr;

        if( strcmp( props->PropName, "ambient_on" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                if( *fptr == 1. )
                    light_switches[AMBIENT] = 1;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "ambient_color" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[AMBIENT].color.value.rgb.red   = *fptr++;
                lights[AMBIENT].color.value.rgb.green = *fptr++;
                lights[AMBIENT].color.value.rgb.blue  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "direct_on" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                if( *fptr == 1. )
                    light_switches[DIRECTIONAL] = 1;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "direct_direction" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[DIRECTIONAL].direction.x = *fptr++;
                lights[DIRECTIONAL].direction.y = *fptr++;
                lights[DIRECTIONAL].direction.z = *fptr++;
                NORMALIZE( lights[DIRECTIONAL].direction );
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "direct_color" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[DIRECTIONAL].color.value.rgb.red   = *fptr++;
                lights[DIRECTIONAL].color.value.rgb.green = *fptr++;
                lights[DIRECTIONAL].color.value.rgb.blue  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "position_on" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                if( *fptr == 1. )
                    light_switches[POSITIONAL] = 1;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "position_point" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[POSITIONAL].point.x = *fptr++;
                lights[POSITIONAL].point.y = *fptr++;
                lights[POSITIONAL].point.z = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "position_atten1" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[POSITIONAL].attenuation1 = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "position_atten2" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[POSITIONAL].attenuation2 = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "position_color" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[POSITIONAL].color.value.rgb.red   = *fptr++;
                lights[POSITIONAL].color.value.rgb.green = *fptr++;
                lights[POSITIONAL].color.value.rgb.blue  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "spot_on" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                if( *fptr == 1. )
                    light_switches[SPOT] = 1;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "spot_point" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[SPOT].point.x = *fptr++;
                lights[SPOT].point.y = *fptr++;
                lights[SPOT].point.z = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "spot_direction" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[SPOT].direction.x = *fptr++;
                lights[SPOT].direction.y = *fptr++;
                lights[SPOT].direction.z = *fptr++;
                NORMALIZE( lights[SPOT].direction );
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "spot_atten1" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[SPOT].attenuation1   = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "spot_atten2" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[SPOT].attenuation2  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "spot_conc" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[SPOT].concentration  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "spot_angle" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[SPOT].spread_angle  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "spot_color" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                lights[SPOT].color.value.rgb.red   = *fptr++;
                lights[SPOT].color.value.rgb.green = *fptr++;
                lights[SPOT].color.value.rgb.blue  = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        props = props->NextProp;
    }

    PEXSetTableEntries( display, rattrs.light_table, 1, 1,
        PEXLUTLight, &lights[AMBIENT] );
    PEXSetTableEntries( display, rattrs.light_table, 2, 1,
        PEXLUTLight, &lights[DIRECTIONAL] );
    PEXSetTableEntries( display, rattrs.light_table, 3, 1,
        PEXLUTLight, &lights[POSITIONAL] );
    PEXSetTableEntries( display, rattrs.light_table, 4, 1,
        PEXLUTLight, &lights[SPOT] );

    numLightsON = numLightsOFF = 0;
    for ( i = 0; i < 4; i++ )
        if ( light_switches[i] )
            lightsON[numLightsON++] = i + 1;
        else
            lightsOFF[numLightsOFF++] = i + 1;

    PEXLabel( display, top_struct, PEXOCStore, LIGHT_STATE_LABEL );
    PEXSetLightSourceState( display, top_struct, PEXOCStore,
        numLightsON, lightsON, numLightsOFF, lightsOFF );
}


void
initialize_depth_cue_table()
{
    PEXDepthCueEntry    depth_cue_rep;

    depth_cue_rep.mode = PEXOn;
    depth_cue_rep.front_plane = 1;
    depth_cue_rep.back_plane = 0;
    depth_cue_rep.front_scaling = 1;
    depth_cue_rep.back_scaling = 0.5;
    depth_cue_rep.color.type = PEXColorTypeRGB;
    SET_COLOR( 0, 1, 0, depth_cue_rep.color.value );
    PEXSetTableEntries( display, rattrs.depth_cue_table, 1, 1,
        PEXLUTDepthCue, &depth_cue_rep );
}


void
initialize_view()
{
    view_ref_pt.x = 0.;          view_up_vec.x = 0.;
    view_ref_pt.y = 0.;          view_up_vec.y = 1.;
    view_ref_pt.z = 0.;          view_up_vec.z = 0.;
    view_plane_normal.x = 1.;    proj_ref_point.x = 0.;
    view_plane_normal.y = 1.;    proj_ref_point.y = 0.;
    view_plane_normal.z = 1.;    proj_ref_point.z = 10.;
    view_window[0].x = -2.0;     view_window[1].x = 2.0;
    view_window[0].y = -2.0;     view_window[1].y = 2.0;
    back_plane  = -10.0;         front_plane = 5.0;
    view_plane  =   0.0;
    viewport.min.x = 0.0;        viewport.max.x = 1.0;
    viewport.min.y = 0.0;        viewport.max.y = 1.0;
    viewport.min.z = 0.0;        viewport.max.z = 1.0;
    view.clip_limits = viewport;
    view.clip_flags = PEXClippingAll;
    perspective = True;

    /* Compute the view orientation matrix. */
    PEXViewOrientationMatrix( &view_ref_pt, &view_plane_normal,
        &view_up_vec, view.orientation );
    PEXViewMappingMatrix( view_window, &viewport, perspective,
        &proj_ref_point, view_plane, back_plane,
        front_plane, view.mapping );
    PEXSetTableEntries( display, rattrs.view_table, 1, 1, PEXLUTView, &view );
    PEXSetTableEntries( display, rattrs.view_table, 2, 1, PEXLUTView, &view );

    view_ref_pt_a.x = 0.;          view_up_vec_a.x = 0.;
    view_ref_pt_a.y = 0.;          view_up_vec_a.y = 1.;
    view_ref_pt_a.z = 0.;          view_up_vec_a.z = 0.;
    view_plane_normal_a.x = 1.;    proj_ref_point_a.x = 0.;
    view_plane_normal_a.y = 1.;    proj_ref_point_a.y = 0.;
    view_plane_normal_a.z = 1.;    proj_ref_point_a.z = 10.;
    view_window_a[0].x = -2.0;     view_window_a[1].x = 2.0;
    view_window_a[0].y = -2.0;     view_window_a[1].y = 2.0;
    back_plane_a  = -10.0;         front_plane_a = 5.0;
    view_plane_a  =   0.0;
    viewport_a.min.x = 0.0;        viewport_a.max.x = 0.15;
    viewport_a.min.y = 0.85;       viewport_a.max.y = 1.0;
    viewport_a.min.z = 0.0;        viewport_a.max.z = 1.0;
    view_a.clip_limits = viewport;
    view_a.clip_flags = PEXClippingAll;
    perspective_a = True;

    /* Compute the view orientation matrix. */
    PEXViewOrientationMatrix( &view_ref_pt_a, &view_plane_normal_a,
        &view_up_vec_a, view_a.orientation );
    PEXViewMappingMatrix( view_window_a, &viewport_a, perspective_a,
        &proj_ref_point_a, view_plane_a, back_plane_a,
        front_plane_a, view_a.mapping );

    /* Entry 3 is used for the new axis view */
    /* Entry 4 is used for erasing the previous axis view */
    PEXSetTableEntries( display, rattrs.view_table, 3, 1, PEXLUTView, &view_a );
    PEXSetTableEntries( display, rattrs.view_table, 4, 1, PEXLUTView, &view_a );
}


void
build_plane()
{
    PEXArrayOfVertex    vertex_data;
    PEXListOfVertex     vertex_data_vert[12];
    PEXListOfVertex     vertex_data_horiz[12];
    PEXArrayOfFacetData facet_data;
    unsigned int        facet_data_flag, vertex_data_flag;
    PEXColorType        colorType;
    unsigned int        num_columns, num_rows;
    register int        i, j, index;
    PEXVertexRGB		*pts, *pts_horiz;

    facet_data_flag = PEXGANone;
    vertex_data_flag = PEXGAColor;
    colorType = PEXColorTypeRGB;
    num_columns = num_rows = 12;

    plane = PEXCreateStructure( display );
    if( verbose )
    fprintf( stderr, "plane\tstructure id = %d\n", plane );

    PEXSetViewIndex( display, plane, PEXOCStore, 2 );
    PEXSetInteriorStyle( display, plane, PEXOCStore, PEXInteriorStyleHollow );
    PEXSetPolylineInterpMethod( display, plane, PEXOCStore, PEXPolylineInterpColor );

    pts       =(PEXVertexRGB*)malloc(num_columns*num_rows*sizeof(PEXVertexRGB));
    pts_horiz =(PEXVertexRGB*)malloc(num_columns*num_rows*sizeof(PEXVertexRGB));

    for (i = 0; i < num_rows; i++) {
        for (j = 0; j < num_columns; j++) {
            index = i * num_rows + j;
            pts[index].point.x = -3.5 + i * 4.4/(float)num_rows;
            pts[index].point.y = -3.5 + j * 4.4/(float)num_columns;
            pts[index].point.z = -3.5;
            if( ((i%2) && !(j%2)) || (!(i%2) && (j%2)) ) {
                pts[index].rgb.red   = 1.0;
                pts[index].rgb.green = 1.0;
                pts[index].rgb.blue  = 1.0;
            }
            else if( (!(i%2) && !(j%2)) || ((i%2) && (j%2)) ) {
                pts[index].rgb.red   = 0.0;
                pts[index].rgb.green = 0.0;
                pts[index].rgb.blue  = 0.0;
            }
        }
        vertex_data_vert[i].count = num_columns;
        vertex_data_vert[i].vertices.rgb = (PEXVertexRGB *)( (char *)pts + 
            i * num_rows * sizeof(PEXVertexRGB) );
    }
    PEXPolylineSetWithData( display, plane, PEXOCStore,
        vertex_data_flag, colorType,
        num_rows, vertex_data_vert );

    for (i = 0; i < num_rows; i++) {
        for (j = 0; j < num_columns; j++) {
            pts_horiz[i*num_rows+j] = pts[j*num_rows+i];
        }
        vertex_data_horiz[i].count = num_rows;
        vertex_data_horiz[i].vertices.rgb = (PEXVertexRGB*)( (char *)pts_horiz +
            i * num_columns * sizeof(PEXVertexRGB) );
    }
    PEXPolylineSetWithData( display, plane, PEXOCStore,
        vertex_data_flag, colorType,
        num_columns, vertex_data_horiz );
#if( 0 ) 
    vertex_data.rgb = pts;
    PEXQuadrilateralMesh( display, plane, PEXOCStore,
        PEXShapeConvex,
        facet_data_flag, vertex_data_flag, colorType,
        facet_data, num_columns, num_rows, vertex_data );
#endif		/*  */

    for (i = 0; i < num_rows; i++) {
        for (j = 0; j < num_columns; j++) {
            index = i * num_rows + j;
            pts[index].point.x = -3.5 + i * 4.4/(float)num_rows;
            pts[index].point.y = -3.5;
            pts[index].point.z = -3.5 + j * 4.4/(float)num_columns;
            if( ((i%2) && !(j%2)) || (!(i%2) && (j%2)) ) {
                pts[index].rgb.red   = 1.0;
                pts[index].rgb.green = 1.0;
                pts[index].rgb.blue  = 1.0;
            }
            else if( (!(i%2) && !(j%2)) || ((i%2) && (j%2)) ) {
                pts[index].rgb.red   = 0.0;
                pts[index].rgb.green = 0.0;
                pts[index].rgb.blue  = 0.0;
            }
        }
        vertex_data_vert[i].count = num_columns;
        vertex_data_vert[i].vertices.rgb = (PEXVertexRGB *)( (char *)pts + 
            i * num_rows * sizeof(PEXVertexRGB) );
    }
    PEXPolylineSetWithData( display, plane, PEXOCStore,
        vertex_data_flag, colorType,
        num_rows, vertex_data_vert );

    for (i = 0; i < num_rows; i++) {
        for (j = 0; j < num_columns; j++) {
            pts_horiz[i*num_rows+j] = pts[j*num_rows+i];
        }
        vertex_data_horiz[i].count = num_rows;
        vertex_data_horiz[i].vertices.rgb = (PEXVertexRGB*)( (char *)pts_horiz +
            i * num_columns * sizeof(PEXVertexRGB) );
    }
    PEXPolylineSetWithData( display, plane, PEXOCStore,
        vertex_data_flag, colorType,
        num_columns, vertex_data_horiz );
#if( 0 ) 
    vertex_data.rgb = pts;
    PEXQuadrilateralMesh( display, plane, PEXOCStore,
        PEXShapeConvex,
        facet_data_flag, vertex_data_flag, colorType,
        facet_data, num_columns, num_rows, vertex_data );
#endif		/*  */

    free( pts );
}


void
build_axis()
{
    PEXColor color;

    static PEXCoord apoints_x[] = { 0.0, 0.0, 0.0, 1.7, 0.0, 0.0 };
    static PEXCoord apoints_y[] = { 0.0, 0.0, 0.0, 0.0, 1.7, 0.0 };
    static PEXCoord apoints_z[] = { 0.0, 0.0, 0.0, 0.0, 0.0, 1.7 };
    static PEXCoord tpoints_x[] = { 1.75, 0.0, 0.0 };
    static PEXCoord tpoints_y[] = { 0.0, 1.75, 0.0 };
    static PEXCoord tpoints_z[] = { 0.0, 0.0, 1.75 };
    static PEXCoord offset;
    static PEXVector vector1, vector2;

    axis = PEXCreateStructure( display );
    if( verbose )
    fprintf( stderr, "axis\tstructure id = %d\n", axis );

#ifdef ANNOTEXT
    PEXSetATextHeight( display, axis, PEXOCStore, .025 );
#else
    PEXSetCharHeight( display, axis, PEXOCStore, .35 );
#endif
    PEXSetLineWidth( display, axis, PEXOCStore, 0.03 );

/*  axis_black */
    PEXSetViewIndex( display, axis, PEXOCStore, 4 );
    SET_COLOR( 0., 0., 0., color );
    PEXSetTextColor( display, axis, PEXOCStore, PEXColorTypeRGB, &color );
    PEXSetLineColor( display, axis, PEXOCStore, PEXColorTypeRGB, &color );

    /* X */
#ifdef ANNOTEXT
    offset.x  = 0.005;  offset.y  = 0.005;  offset.z  = 0.005;
    PEXAnnotationText( display, axis, PEXOCStore,
        &apoints_x[1], &offset, 1, "X" );
#else
    vector1.x = 1.000;  vector1.y = 0.000;  vector1.z = 0.000;
    vector2.x = 0.000;  vector2.y = 1.000;  vector2.z = 0.000;
    PEXText( display, axis, PEXOCStore,
        tpoints_x, &vector1, &vector2, 1, "X" );
#endif
    PEXPolyline( display, axis, PEXOCStore, 2, apoints_x );

    /* Y */
#ifdef ANNOTEXT
    offset.x  = 0.005;  offset.y  = 0.005;  offset.z  = 0.005;
    PEXAnnotationText( display, axis, PEXOCStore,
        &apoints_y[1], &offset, 1, "Y" );
#else
    vector1.x = 0.670;  vector1.y = 0.670;  vector1.z = 0.670;
    vector2.x = 0.000;  vector2.y = 1.000;  vector2.z = 0.000;
    PEXText( display, axis, PEXOCStore,
        tpoints_y, &vector1, &vector2, 1, "Y" );
#endif
    PEXPolyline( display, axis, PEXOCStore, 2, apoints_y );

    /* Z */
#ifdef ANNOTEXT
    offset.x = -0.035;  offset.y  = -0.02;  offset.z = -0.035;
    PEXAnnotationText( display, axis, PEXOCStore,
        &apoints_z[1], &offset, 1, "Z" );
#else
    vector1.x = 0.000;  vector1.y = 1.000;  vector1.z = 0.000;
    vector2.x = 0.000;  vector2.y = 0.000;  vector2.z = 1.000;
    PEXText( display, axis, PEXOCStore,
        tpoints_z, &vector1, &vector2, 1, "Z" );
#endif
    PEXPolyline( display, axis, PEXOCStore, 2, apoints_z );

/*  axis */
    PEXSetViewIndex( display, axis, PEXOCStore, 3 );
    PEXSetLineWidth( display, axis, PEXOCStore, 0. );

    /* X */
    SET_COLOR( 1., 0., 0., color );
    PEXSetTextColor( display, axis, PEXOCStore, PEXColorTypeRGB, &color );
#ifdef ANNOTEXT
    offset.x  = 0.005;  offset.y  = 0.005;  offset.z  = 0.005;
    PEXAnnotationText( display, axis, PEXOCStore,
        &apoints_x[1], &offset, 1, "X" );
#else
    vector1.x = 1.000;  vector1.y = 0.000;  vector1.z = 0.000;
    vector2.x = 0.000;  vector2.y = 1.000;  vector2.z = 0.000;
    PEXText( display, axis, PEXOCStore,
        tpoints_x, &vector1, &vector2, 1, "X" );
#endif
    PEXSetLineColor( display, axis, PEXOCStore, PEXColorTypeRGB, &color );
    PEXPolyline( display, axis, PEXOCStore, 2, apoints_x );

    /* Y */
    SET_COLOR( 0., 1., 0., color );
    PEXSetTextColor( display, axis, PEXOCStore, PEXColorTypeRGB, &color );
#ifdef ANNOTEXT
    offset.x  = 0.005;  offset.y  = 0.005;  offset.z  = 0.005;
    PEXAnnotationText( display, axis, PEXOCStore,
        &apoints_y[1], &offset, 1, "Y" );
#else
    vector1.x = 0.670;  vector1.y = 0.670;  vector1.z = 0.670;
    vector2.x = 0.000;  vector2.y = 1.000;  vector2.z = 0.000;
    PEXText( display, axis, PEXOCStore,
        tpoints_y, &vector1, &vector2, 1, "Y" );
#endif
    PEXSetLineColor( display, axis, PEXOCStore, PEXColorTypeRGB, &color );
    PEXPolyline( display, axis, PEXOCStore, 2, apoints_y );

    /* Z */
    SET_COLOR( 0., 0., 1., color );
    PEXSetTextColor( display, axis, PEXOCStore, PEXColorTypeRGB, &color );
#ifdef ANNOTEXT
    offset.x = -0.035;  offset.y  = -0.02;  offset.z = -0.035;
    PEXAnnotationText( display, axis, PEXOCStore,
        &apoints_z[1], &offset, 1, "Z" );
#else
    vector1.x = 0.000;  vector1.y = 1.000;  vector1.z = 0.000;
    vector2.x = 0.000;  vector2.y = 0.000;  vector2.z = 1.000;
    PEXText( display, axis, PEXOCStore,
        tpoints_z, &vector1, &vector2, 1, "Z" );
#endif
    PEXSetLineColor( display, axis, PEXOCStore, PEXColorTypeRGB, &color );
    PEXPolyline( display, axis, PEXOCStore, 2, apoints_z );
}


void
build_object_shadow( Obj )
OFFObjDesc *Obj;
{
    PEXListOfVertex    *vertex_data;
    PEXFacetData       *facet_data;

    int                 shape_hint, ignore_edges, contour_hint, color_type;
    unsigned int        facet_attributes, vertex_attributes;
    double              theta = 0, phi = 0;
    register int        i, j, index;

    OFFProperty *props;

    long nver, nfac, nedg;
    unsigned int nedges;
    long n_vertex_normals, n_facet_normals;
    long n_vertex_colors, n_facet_colors;
    int  vertex_order = COUNTERCLOCKWISE;

    Vertex      *vertex_normals = NULL;
    PEXColorRGB *vertex_colors  = NULL;

    Vertex      *facet_normals = NULL;
    PEXColorRGB *facet_colors  = NULL;

    shape_hint   = PEXShapeConvex;
    ignore_edges = PEXFalse;
    contour_hint = PEXContourDisjoint;
    facet_attributes  = PEXGANone;
    vertex_attributes = PEXGANone;
    color_type = PEXColorTypeRGB;

    object_shadow = PEXCreateStructure( display );
    if( verbose )
    fprintf( stderr, "shadow\tstructure id = %d\n", object_shadow );

    props = Obj->FirstProp;
    while( props != NULL ) {

        char  *ptr;    long  *lptr;    float *fptr;    short *sptr;

        if( strcmp( props->PropName, "connect_offset" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                connect_offset = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "geometry" ) == 0 ) {
            if( props->PropType == OFF_INDEXED_POLY_DATA ) {
                ptr  = props->PropData;
                lptr = (long *) ptr;
                nver = *lptr++;
                nfac = *lptr++;
                nedg = *lptr++;
                ptr  = (char *) lptr;

                nvertices = nver;
                nfacets   = nfac;
                nedges    = nedg;

                vertices = (Vertex *) malloc ( nvertices * sizeof( Vertex ) );
                facets   = (Facet *) malloc ( nfacets * sizeof( Facet ) );

                fptr = (float *) ptr;
                for(i = 0; i < nvertices; i++) {
                    vertices[i].x = *fptr++;
                    vertices[i].y = *fptr++;
                    vertices[i].z = *fptr++;
                }
#ifdef DEBUG
                for(i = 0; i < nvertices; i++)
                    printf( "%f %f %f\n", vertices[i].x, vertices[i].y, vertices[i].z );
#endif    /*  */
                ptr = (char *) fptr;

                sptr = (short *) ptr;
                for(j = 0; j < nfacets; j++) {
                    facets[j].count = *sptr++;
                    facets[j].edges = (Edge *) malloc ( facets[j].count * sizeof( Edge ) );
                }
                for(j = 0; j < nfacets; j++)
                    for(i = 0; i < facets[j].count; i++)
                        facets[j].edges[i].e = *sptr++;
#ifdef DEBUG
                for(j = 0; j < nfacets; j++) {
                    printf( "%d ", facets[j].count );
                    for(i = 0; i < facets[j].count; i++)
                        printf( "%d ", facets[j].edges[i].e );
                    printf( "\n" );
                }
                printf( "\n" );
#endif    /*  */
                ptr = (char *) sptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "vertex_order" ) == 0 )
            vertex_order = (!strcmp( *((char **) props->PropData), "clockwise"))
                         ? CLOCKWISE : COUNTERCLOCKWISE;

        else if( strcmp( props->PropName, "rotation" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {

                n_rotation = 1;

                ptr = props->PropData;

                fptr = (float *) ptr;

                rot_x = *fptr++;
                rot_y = *fptr++;
                rot_z = *fptr++;

                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "translation" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {

                n_translation = 1;

                ptr = props->PropData;

                fptr = (float *) ptr;

                translation.x = *fptr++;
                translation.y = *fptr++;
                translation.z = *fptr++;

                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "scale" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {

                n_scale = 1;

                ptr = props->PropData;

                fptr = (float *) ptr;

                scale.x = *fptr++;
                scale.y = *fptr++;
                scale.z = *fptr++;

                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        props = props->NextProp;
    }

    PEXIdentityMatrix( shad_local_matrix );
    if( n_rotation ) {
        PEXMatrix m1, m2, m3, new_matrix;
        PEXRotate( PEXXAxis, DEGRAD(rot_x), m1 );
        PEXRotate( PEXYAxis, DEGRAD(rot_y), m2 );
        PEXRotate( PEXZAxis, DEGRAD(rot_z), m3 );
        PEXMatrixMult( m1, m2, new_matrix );
        PEXMatrixMult( new_matrix, m3, new_matrix );
        PEXMatrixMult( shad_local_matrix, new_matrix, shad_local_matrix );
    }
    if( n_translation ) {
        PEXMatrix new_matrix;
        PEXTranslate( &translation, new_matrix );
        PEXMatrixMult( shad_local_matrix, new_matrix, shad_local_matrix );
    }
    if( n_scale ) {
        PEXMatrix new_matrix;
        PEXScale( &scale, new_matrix );
        PEXMatrixMult( shad_local_matrix, new_matrix, shad_local_matrix );
    }
    {
        PEXCoord4D    light;
        PEXMatrix    new_matrix;
        light.x = 10.0;
        light.y = 100.0;
        light.z = 7.0;
        light.w = PEXLightInfinite;
        /* This should have been -3.49, but it doesnt look good */
        PEXUmbraMatrix( PEXYPlane, -2.49, light, new_matrix );
        PEXMatrixMult( shad_local_matrix, new_matrix, shad_local_matrix );
    }
    PEXLabel( display, object_shadow, PEXOCStore, SHADOW_ROTATION_LABEL );
    PEXSetLocalTransform( display, object_shadow, PEXOCStore,
        PEXReplace, shad_local_matrix );

    SET_COLOR( 0.35, 0.35, 0.35, colorShadow );
    PEXSetSurfaceColor( display, object_shadow, PEXOCStore,
        PEXColorTypeRGB, &colorShadow );
    PEXSetBFSurfaceColor( display, object_shadow, PEXOCStore,
        PEXColorTypeRGB, &colorShadow );

    facet_data  = (PEXFacetData *)   malloc( nfacets * sizeof(PEXFacetData) );
    vertex_data = (PEXListOfVertex *)malloc( nfacets * sizeof(PEXListOfVertex));

    if( vertex_order == CLOCKWISE )
    for(j = 0; j < nfacets; j++) {

        vertex_data[j].count = facets[j].count;

        {
            PEXCoord *vtx;
            vtx = (PEXCoord *)malloc( vertex_data[j].count *
                                      sizeof(PEXCoord) );
            vertex_data[j].vertices.no_data = vtx;
            for(i = vertex_data[j].count-1; i >= 0; i--, vtx++) {
                vtx->x = vertices[facets[j].edges[i].e - connect_offset].x;
                vtx->y = vertices[facets[j].edges[i].e - connect_offset].y;
                vtx->z = vertices[facets[j].edges[i].e - connect_offset].z;
            }
        }

        PEXFillAreaWithData( display, object_shadow, PEXOCStore,
            shape_hint, ignore_edges,
            facet_attributes, vertex_attributes,
            color_type,
            &facet_data[j],
            vertex_data[j].count, vertex_data[j].vertices );
    }

    else /* COUNTERCLOCKWISE */
    for(j = 0; j < nfacets; j++) {

        vertex_data[j].count = facets[j].count;

        {
            PEXCoord *vtx;
            vtx = (PEXCoord *)malloc( vertex_data[j].count *
                                      sizeof(PEXCoord) );
            vertex_data[j].vertices.no_data = vtx;
            for(i = 0; i < vertex_data[j].count; i++, vtx++) {
                vtx->x = vertices[facets[j].edges[i].e - connect_offset].x;
                vtx->y = vertices[facets[j].edges[i].e - connect_offset].y;
                vtx->z = vertices[facets[j].edges[i].e - connect_offset].z;
            }
        }

        PEXFillAreaWithData( display, object_shadow, PEXOCStore,
            shape_hint, ignore_edges,
            facet_attributes, vertex_attributes,
            color_type,
            &facet_data[j],
            vertex_data[j].count, vertex_data[j].vertices );
    }

    free( facet_data );
    for(j = 0; j < nfacets; j++) {
        free( vertex_data[j].vertices.no_data );
    }
    free( vertex_data );
}


void
build_object( Obj )
OFFObjDesc *Obj;
{
    PEXListOfVertex *vertex_data;
    PEXFacetData    *facet_data;

    PEXArrayOfVertex     vertex_sofas;
    PEXArrayOfFacetData  facet_sofas;
    PEXConnectivityData *conn_sofas;

    double theta = 0, phi = 0;
    register int i, j, k, index;

    OFFProperty *props;

    unsigned int edge_attributes;

    long nver, nfac, nedg;
    unsigned int nedges;
    long n_vertex_normals, n_facet_normals;
    long n_vertex_colors, n_facet_colors;
    int  vertex_order = COUNTERCLOCKWISE;

    Vertex      *vertex_normals = NULL;
    PEXColorRGB *vertex_colors  = NULL;

    Vertex      *facet_normals = NULL;
/*  PEXColorRGB *facet_colors  = NULL; */

    shape_hint   = PEXShapeConvex;
    ignore_edges = PEXFalse;
    contour_hint = PEXContourDisjoint;
    facet_attributes  = PEXGANone;
    vertex_attributes = PEXGANone;
    edge_attributes   = PEXGANone;
    color_type = PEXColorTypeRGB;

    object = PEXCreateStructure( display );
    if( verbose )
    fprintf( stderr, "object\tstructure id = %d\n", object );

    props = Obj->FirstProp;
    while( props != NULL ) {

        char  *ptr;    long  *lptr;    float *fptr;    short *sptr;

        if( strcmp( props->PropName, "connect_offset" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;
                fptr = (float *) ptr;
                connect_offset = *fptr++;
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "geometry" ) == 0 ) {
            if( props->PropType == OFF_INDEXED_POLY_DATA ) {
                ptr  = props->PropData;
                lptr = (long *) ptr;
                nver = *lptr++;
                nfac = *lptr++;
                nedg = *lptr++;
                ptr  = (char *) lptr;

                nvertices = nver;
                nfacets   = nfac;
                nedges    = nedg;

                vertices = (Vertex *) malloc ( nvertices * sizeof( Vertex ) );
                facets   = (Facet *) malloc ( nfacets * sizeof( Facet ) );

                fptr = (float *) ptr;
                for(i = 0; i < nvertices; i++) {
                    vertices[i].x = *fptr++;
                    vertices[i].y = *fptr++;
                    vertices[i].z = *fptr++;
                }
#ifdef DEBUG
                for(i = 0; i < nvertices; i++)
                    printf( "%f %f %f\n", vertices[i].x, vertices[i].y, vertices[i].z );
#endif    /*  */
                ptr = (char *) fptr;

                sptr = (short *) ptr;
                for(j = 0; j < nfacets; j++) {
                    facets[j].count = *sptr++;
                    facets[j].edges = (Edge *) malloc ( facets[j].count * sizeof( Edge ) );
                }
                for(j = 0; j < nfacets; j++)
                    for(i = 0; i < facets[j].count; i++)
                        facets[j].edges[i].e = *sptr++;
#ifdef DEBUG
                for(j = 0; j < nfacets; j++) {
                    printf( "%d ", facets[j].count );
                    for(i = 0; i < facets[j].count; i++)
                        printf( "%d ", facets[j].edges[i].e );
                    printf( "\n" );
                }
                printf( "\n" );
#endif    /*  */
                ptr = (char *) sptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "vertex_normals" ) == 0 ) {
            if( props->PropType == OFF_GENERIC_DATA ) {
                ptr = props->PropData;
                lptr = (long *) ptr;
                n_vertex_normals = *lptr++;
                ptr = (char *) lptr;

                vertex_normals = (Vertex *) malloc ( n_vertex_normals *
                                                     sizeof( Vertex ) );
                fptr = (float *) ptr;
                for(i = 0; i < n_vertex_normals; i++) {
                    vertex_normals[i].x = *fptr++;
                    vertex_normals[i].y = *fptr++;
                    vertex_normals[i].z = *fptr++;
                }
#ifdef DEBUG
                for(i = 0; i < n_vertex_normals; i++)
                    printf( "%f %f %f\n", vertex_normals[i].x,
                                          vertex_normals[i].y,
                                          vertex_normals[i].z );
                printf( "\n" );
#endif    /*  */
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "polygon_normals" ) == 0 ) {
            if( props->PropType == OFF_GENERIC_DATA ) {
                ptr = props->PropData;
                lptr = (long *) ptr;
                n_facet_normals = *lptr++;
                ptr = (char *) lptr;

                facet_normals = (Vertex *) malloc ( n_facet_normals *
                                                    sizeof( Vertex ) );
                fptr = (float *) ptr;
                for(i = 0; i < n_facet_normals; i++) {
                    facet_normals[i].x = *fptr++;
                    facet_normals[i].y = *fptr++;
                    facet_normals[i].z = *fptr++;
                }
#ifdef DEBUG
                for(i = 0; i < n_facet_normals; i++) {
                    printf( "%f %f %f\n", facet_normals[i].x,facet_normals[i].y,
                                          facet_normals[i].z );
                }
                printf( "\n" );
#endif    /*  */
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "vertex_colors" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;

                /*
                   ASSUMPTION: variable nvertices has been set before this 
                   execution this can only happen if geometry is defined 
                   before vertex_colors
                */
                n_vertex_colors = nvertices;

                vertex_colors = (PEXColorRGB *) malloc( n_vertex_colors *
                                                    sizeof( PEXColorRGB ) );
                fptr = (float *) ptr;

                vertex_colors[0].red   = *fptr++;
                vertex_colors[0].green = *fptr++;
                vertex_colors[0].blue  = *fptr++;
                for(i = 1; i < n_vertex_colors; i++) {
                    vertex_colors[i].red   = vertex_colors[0].red;
                    vertex_colors[i].green = vertex_colors[0].green;
                    vertex_colors[i].blue  = vertex_colors[0].blue;
                }

                ptr = (char *) fptr;
            }

            else if( props->PropType == OFF_GENERIC_DATA ) {
                ptr = props->PropData;
                lptr = (long *) ptr;
                n_vertex_colors = *lptr++;

                vertex_colors = (PEXColorRGB *) malloc ( n_vertex_colors *
                                                    sizeof( PEXColorRGB ) );
                fptr = (float *) lptr;
                for(i = 0; i < n_vertex_colors; i++) {
                    vertex_colors[i].red  = *fptr++;
                    vertex_colors[i].green= *fptr++;
                    vertex_colors[i].blue = *fptr++;
                }
                ptr = (char *) fptr;
            }
            else
            {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "polygon_colors" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {
                ptr = props->PropData;

                /*
                   ASSUMPTION: variable nfacets has been set before this 
                   execution this can only happen if geometry is defined 
                   before polygon_colors
                */
                n_facet_colors = nfacets;

                facet_colors = (PEXColorRGB *) malloc( n_facet_colors *
                                                    sizeof( PEXColorRGB ) );
                fptr = (float *) ptr;

                facet_colors[0].red   = *fptr++;
                facet_colors[0].green = *fptr++;
                facet_colors[0].blue  = *fptr++;
                for(i = 1; i < n_facet_colors; i++) {
                    facet_colors[i].red   = facet_colors[0].red;
                    facet_colors[i].green = facet_colors[0].green;
                    facet_colors[i].blue  = facet_colors[0].blue;
                }

                ptr = (char *) fptr;
            }
            else if( props->PropType == OFF_GENERIC_DATA ) {
                ptr = props->PropData;
                lptr = (long *) ptr;
                n_facet_colors = *lptr++;
                ptr = (char *) lptr;

                facet_colors = (PEXColorRGB *) malloc( n_facet_colors *
                                                    sizeof( PEXColorRGB ) );
                fptr = (float *) ptr;

                for(i = 0; i < n_facet_colors; i++) {
                    facet_colors[i].red = *fptr++;
                    facet_colors[i].green = *fptr++;
                    facet_colors[i].blue = *fptr++;
                }

                ptr = (char *) fptr;
            }
            else if( props->PropType == OFF_INDEXED_DATA ) {
                long n_colors;
                PEXColorRGB *colors;

                ptr = props->PropData;
                lptr =( long *) ptr;
                n_colors = *lptr++;
                n_facet_colors = *lptr++;

                colors = (PEXColorRGB *) malloc( n_colors *
                                                sizeof( PEXColorRGB ) );
                facet_colors = (PEXColorRGB *) malloc( n_facet_colors *
                                                    sizeof( PEXColorRGB ) );
                fptr = (float *) lptr;

                for(i = 0; i < n_colors; i++) {
                    /* use vertex buffers for temp storage */
                    colors[i].red  = *fptr++;
                    colors[i].green= *fptr++;
                    colors[i].blue = *fptr++;
                }
                sptr = (short *) fptr;

                for(i = 0; i < n_facet_colors; i++) {
                    j = *sptr++ - 1;
                    facet_colors[i].red  = colors[j].red;
                    facet_colors[i].green= colors[j].green;
                    facet_colors[i].blue = colors[j].blue;
                }

                ptr = (char *) sptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "vertex_order" ) == 0 )
            vertex_order = (!strcmp( *((char **) props->PropData), "clockwise"))
                         ? CLOCKWISE : COUNTERCLOCKWISE;

        else if( strcmp( props->PropName, "flyby_prp" ) == 0 ) {
            if( props->PropType == OFF_GENERIC_DATA ) {
                ptr = props->PropData;
                lptr = (long *) ptr;
                n_prp_points = *lptr++;
                ptr = (char *) lptr;

                prp_points = (PEXVector *) malloc( n_prp_points *
                                                    sizeof( PEXVector ) );
                fptr = (float *) ptr;

                for(i = 0; i < n_prp_points; i++) {
                    prp_points[i].x = *fptr++;
                    prp_points[i].y = *fptr++;
                    prp_points[i].z = *fptr++;
                }

                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "flyby_vpn" ) == 0 ) {
            if( props->PropType == OFF_GENERIC_DATA ) {
                ptr = props->PropData;
                lptr = (long *) ptr;
                n_vpn_points = *lptr++;
                ptr = (char *) lptr;

                vpn_points = (PEXVector *) malloc( n_vpn_points *
                                                    sizeof( PEXVector ) );
                fptr = (float *) ptr;

                for(i = 0; i < n_vpn_points; i++) {
                    vpn_points[i].x = *fptr++;
                    vpn_points[i].y = *fptr++;
                    vpn_points[i].z = *fptr++;
                }

                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "rotation" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {

                n_rotation = 1;

                ptr = props->PropData;

                fptr = (float *) ptr;

                rot_x = *fptr++;
                rot_y = *fptr++;
                rot_z = *fptr++;

                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "translation" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {

                n_translation = 1;

                ptr = props->PropData;

                fptr = (float *) ptr;

                translation.x = *fptr++;
                translation.y = *fptr++;
                translation.z = *fptr++;

                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "scale" ) == 0 ) {
            if( props->PropType == OFF_DEFAULT_DATA ) {

                n_scale = 1;

                ptr = props->PropData;

                fptr = (float *) ptr;

                scale.x = *fptr++;
                scale.y = *fptr++;
                scale.z = *fptr++;

                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        else if( strcmp( props->PropName, "morph_vertices" ) == 0 ) {
            if( props->PropType == OFF_INDEXED_POLY_DATA ) {
                ptr = props->PropData;
                lptr = (long *) ptr;
                nver = *lptr++;
                nfac = *lptr++;
                nedges = *lptr++;
                ptr = (char *) lptr;

                nvertices = nver;
                nfacets   = nfac;

                morph_vertices = (Vertex *) malloc( nvertices*sizeof(Vertex) );

                fptr = (float *) ptr;
                for(i = 0; i < nvertices; i++) {
                    morph_vertices[i].x = *fptr++;
                    morph_vertices[i].y = *fptr++;
                    morph_vertices[i].z = *fptr++;
                }
#ifdef DEBUG
                for(i = 0; i < nvertices; i++)
                    printf( "%f %f %f\n", morph_vertices[i].x, morph_vertices[i].y, morph_vertices[i].z );
#endif    /*  */
                ptr = (char *) fptr;
            }
            else {
                fprintf ( stderr, "Unsupported type\n" );
                fprintf ( stderr, "Property Name: %s\n", props->PropName );
                fprintf ( stderr, "Property Type: %d\n", props->PropType );
            }
        }

        props = props->NextProp;
    }

    PEXIdentityMatrix( local_matrix );
    if( n_rotation ) {
        PEXMatrix m1, m2, m3, new_matrix;
        PEXRotate( PEXXAxis, DEGRAD(rot_x), m1 );
        PEXRotate( PEXYAxis, DEGRAD(rot_y), m2 );
        PEXRotate( PEXZAxis, DEGRAD(rot_z), m3 );
        PEXMatrixMult( m1, m2, new_matrix );
        PEXMatrixMult( new_matrix, m3, new_matrix );
        PEXMatrixMult( local_matrix, new_matrix, local_matrix );
    }
    if( n_translation ) {
        PEXMatrix new_matrix;
        PEXTranslate( &translation, new_matrix );
        PEXMatrixMult( local_matrix, new_matrix, local_matrix );
    }
    if( n_scale ) {
        PEXMatrix new_matrix;
        PEXScale( &scale, new_matrix );
        PEXMatrixMult( local_matrix, new_matrix, local_matrix );
    }
    PEXLabel( display, object, PEXOCStore, ROTATION_LABEL );
    PEXSetLocalTransform( display, object, PEXOCStore,
        PEXReplace, local_matrix );

    for ( i=0; i<4; i++ )
    for ( j=0; j<4; j++ )
        init_matrix[i][j] = local_matrix[i][j];

    facet_data  = (PEXFacetData *)malloc( nfacets * sizeof(PEXFacetData) );
    vertex_data = (PEXListOfVertex *)malloc( nfacets * sizeof(PEXListOfVertex));
    conn_sofas  = (PEXConnectivityData *)malloc( nfacets *
                                                 sizeof(PEXConnectivityData) );

    if( facet_normals  != NULL ) facet_attributes  |= PEXGANormal;
    if( facet_colors   != NULL ) facet_attributes  |= PEXGAColor;
    if( vertex_normals != NULL ) vertex_attributes |= PEXGANormal;
    if( vertex_colors  != NULL ) vertex_attributes |= PEXGAColor;
#ifdef SOFAS
    if( (facet_attributes & PEXGANormal) &&
        (facet_attributes & PEXGAColor) ) {
        PEXColorRGBNormal *fctrn_sofas = (PEXColorRGBNormal *)malloc( nfacets *
                                         sizeof( PEXColorRGBNormal ) );
        facet_sofas.rgb_normal  = fctrn_sofas;
        fctrn_sofas[j].normal.x = facet_normals[j].x;
        fctrn_sofas[j].normal.y = facet_normals[j].y;
        fctrn_sofas[j].normal.z = facet_normals[j].z;
        fctrn_sofas[j].rgb.red  = facet_colors[j].red;
        fctrn_sofas[j].rgb.green= facet_colors[j].green;
        fctrn_sofas[j].rgb.blue = facet_colors[j].blue;
    }
    else if( facet_attributes & PEXGANormal ) {
        PEXVector *fctn_sofas = (PEXVector *)malloc( nfacets *
                                sizeof( PEXVector ) );
        facet_sofas.normal = fctn_sofas;
        fctn_sofas[j].x = facet_normals[j].x;
        fctn_sofas[j].y = facet_normals[j].y;
        fctn_sofas[j].z = facet_normals[j].z;
    }
    else if( facet_attributes & PEXGAColor ) {
        PEXColorRGB *fctr_sofas = (PEXColorRGB *)malloc( nfacets *
                                  sizeof( PEXColorRGB ) );
        facet_sofas.rgb = fctr_sofas;
        fctr_sofas[j].red  = facet_colors[j].red;
        fctr_sofas[j].green= facet_colors[j].green;
        fctr_sofas[j].blue = facet_colors[j].blue;
    }

    if( (vertex_attributes & PEXGANormal) &&
        (vertex_attributes & PEXGAColor) ) {
        PEXVertexRGBNormal *vtxrn_sofas = (PEXVertexRGBNormal *)
                            malloc( nvertices * sizeof( PEXVertexRGBNormal ) );
        vertex_sofas.rgb_normal = vtxrn_sofas;
        for(i = 0; i < nvertices; i++ ) {
           vtxrn_sofas[i].point.x   = vertices[i].x;
           vtxrn_sofas[i].point.y   = vertices[i].y;
           vtxrn_sofas[i].point.z   = vertices[i].z;
           vtxrn_sofas[i].rgb.red   = vertex_colors[i].red;
           vtxrn_sofas[i].rgb.green = vertex_colors[i].green;
           vtxrn_sofas[i].rgb.blue  = vertex_colors[i].blue;
           vtxrn_sofas[i].normal.x  = vertex_normals[i].x;
           vtxrn_sofas[i].normal.y  = vertex_normals[i].y;
           vtxrn_sofas[i].normal.z  = vertex_normals[i].z;
        }
    }
    else if( vertex_attributes & PEXGAColor ) {
        PEXVertexRGB *vtxr_sofas = (PEXVertexRGB *) malloc( nvertices *
                                   sizeof( PEXVertexRGB ) );
        vertex_sofas.rgb = vtxr_sofas;
        for(i = 0; i < nvertices; i++ ) {
           vtxr_sofas[i].point.x   = vertices[i].x;
           vtxr_sofas[i].point.y   = vertices[i].y;
           vtxr_sofas[i].point.z   = vertices[i].z;
           vtxr_sofas[i].rgb.red   = vertex_colors[i].red;
           vtxr_sofas[i].rgb.green = vertex_colors[i].green;
           vtxr_sofas[i].rgb.blue  = vertex_colors[i].blue;
        }
    }
    else if( vertex_attributes & PEXGANormal ) {
        PEXVertexNormal *vtxn_sofas = (PEXVertexNormal *) malloc( nvertices *
                                      sizeof( PEXVertexNormal ));
        vertex_sofas.normal = vtxn_sofas;
        for(i = 0; i < nvertices; i++ ) {
           vtxn_sofas[i].point.x   = vertices[i].x;
           vtxn_sofas[i].point.y   = vertices[i].y;
           vtxn_sofas[i].point.z   = vertices[i].z;
           vtxn_sofas[i].normal.x  = vertex_normals[i].x;
           vtxn_sofas[i].normal.y  = vertex_normals[i].y;
           vtxn_sofas[i].normal.z  = vertex_normals[i].z;
        }
    }
    else {
        PEXCoord *vtx_sofas = (PEXCoord *) malloc( nvertices *
                              sizeof( PEXCoord ) );
        vertex_sofas.no_data = vtx_sofas;
        for(i = 0; i < nvertices; i++ ) {
           vtx_sofas[i].x = vertices[i].x;
           vtx_sofas[i].y = vertices[i].y;
           vtx_sofas[i].z = vertices[i].z;
        }
    }

    for( j = 0; j < nfacets; j++ ) {
        conn_sofas[j].count = 1;
        conn_sofas[j].lists = (PEXListOfUShort *)malloc( 
                                                 sizeof( PEXListOfUShort ) );
        conn_sofas[j].lists->count  = facets[j].count;
        conn_sofas[j].lists->shorts = (unsigned short *)malloc(
                        conn_sofas[j].lists->count * sizeof(unsigned short) );
        for( k = 0; k < conn_sofas[j].lists->count; k++ )
        conn_sofas[j].lists->shorts[k] = facets[j].edges[k].e - connect_offset;
    }

    PEXSetOfFillAreaSets( display, object, PEXOCStore,
        shape_hint, facet_attributes, vertex_attributes, edge_attributes,
        contour_hint, True, color_type,
        nfacets, facet_sofas, nvertices, vertex_sofas,
        nedges, (PEXSwitch *)NULL, conn_sofas );

    if( (facet_attributes & PEXGANormal) &&
        (facet_attributes & PEXGAColor) )
         free( facet_sofas.rgb_normal );
    else if( facet_attributes & PEXGANormal )
         free( facet_sofas.normal );
    else if( facet_attributes & PEXGAColor )
         free( facet_sofas.rgb );

    if( (vertex_attributes & PEXGANormal) &&
        (vertex_attributes & PEXGAColor) )
         free( vertex_sofas.rgb_normal );
    else if( vertex_attributes & PEXGANormal )
         free( vertex_sofas.normal );
    else if( vertex_attributes & PEXGAColor )
         free( vertex_sofas.rgb );
    else
         free( vertex_sofas.no_data );

    for( j = 0; j < nfacets; j++ ) {
        free( conn_sofas[j].lists->shorts );
        free( conn_sofas[j].lists );
    }
    free( conn_sofas );
#endif		/*  */
    if( vertex_order == CLOCKWISE ) {
    for(j = 0; j < nfacets; j++) {

        if( (facet_attributes & PEXGANormal) &&
            (facet_attributes & PEXGAColor) ) {
            facet_data[j].rgb_normal.normal.x = facet_normals[j].x;
            facet_data[j].rgb_normal.normal.y = facet_normals[j].y;
            facet_data[j].rgb_normal.normal.z = facet_normals[j].z;
            facet_data[j].rgb_normal.rgb.red  = facet_colors[j].red;
            facet_data[j].rgb_normal.rgb.green= facet_colors[j].green;
            facet_data[j].rgb_normal.rgb.blue = facet_colors[j].blue;
        }
        else if( facet_attributes & PEXGANormal ) {
            facet_data[j].normal.x = facet_normals[j].x;
            facet_data[j].normal.y = facet_normals[j].y;
            facet_data[j].normal.z = facet_normals[j].z;
        }
        else if( facet_attributes & PEXGAColor ) {
            facet_data[j].rgb.red  = facet_colors[j].red;
            facet_data[j].rgb.green= facet_colors[j].green;
            facet_data[j].rgb.blue = facet_colors[j].blue;
        }

        vertex_data[j].count = facets[j].count;

        if( (vertex_attributes & PEXGANormal) &&
            (vertex_attributes & PEXGAColor) ) {
            PEXVertexRGBNormal *vtxrn;

            vtxrn = (PEXVertexRGBNormal *) malloc( vertex_data[j].count *
                                                 sizeof( PEXVertexRGBNormal ) );
            vertex_data[j].vertices.rgb_normal = vtxrn;
            for(i = vertex_data[j].count-1; i >= 0; i--, vtxrn++) {
                vtxrn->point.x  = vertices[facets[j].edges[i].e - connect_offset].x;
                vtxrn->point.y  = vertices[facets[j].edges[i].e - connect_offset].y;
                vtxrn->point.z  = vertices[facets[j].edges[i].e - connect_offset].z;
                vtxrn->rgb.red  = vertex_colors[facets[j].edges[i].e - connect_offset].red;
                vtxrn->rgb.green= vertex_colors[facets[j].edges[i].e - connect_offset].green;
                vtxrn->rgb.blue = vertex_colors[facets[j].edges[i].e - connect_offset].blue;
                vtxrn->normal.x = vertex_normals[facets[j].edges[i].e - connect_offset].x;
                vtxrn->normal.y = vertex_normals[facets[j].edges[i].e - connect_offset].y;
                vtxrn->normal.z = vertex_normals[facets[j].edges[i].e - connect_offset].z;
            }
        }
        else if( vertex_attributes & PEXGAColor ) {
            PEXVertexRGB *vtxr;

            vtxr = (PEXVertexRGB *) malloc( vertex_data[j].count *
                                               sizeof( PEXVertexRGB ) ) ;
            vertex_data[j].vertices.rgb = vtxr;
            for(i = vertex_data[j].count-1; i >= 0; i--, vtxr++) {
                vtxr->point.x  = vertices[facets[j].edges[i].e - connect_offset].x;
                vtxr->point.y  = vertices[facets[j].edges[i].e - connect_offset].y;
                vtxr->point.z  = vertices[facets[j].edges[i].e - connect_offset].z;
                vtxr->rgb.red  = vertex_colors[facets[j].edges[i].e - connect_offset].red;
                vtxr->rgb.green= vertex_colors[facets[j].edges[i].e - connect_offset].green;
                vtxr->rgb.blue = vertex_colors[facets[j].edges[i].e - connect_offset].blue;
            }
        }
        else if( vertex_attributes & PEXGANormal ) {
            PEXVertexNormal *vtxn;

            vtxn = (PEXVertexNormal *) malloc( vertex_data[j].count *
                                               sizeof( PEXVertexNormal ) ) ;
            vertex_data[j].vertices.normal = vtxn;
            for(i = vertex_data[j].count-1; i >= 0; i--, vtxn++) {
                vtxn->point.x  = vertices[facets[j].edges[i].e - connect_offset].x;
                vtxn->point.y  = vertices[facets[j].edges[i].e - connect_offset].y;
                vtxn->point.z  = vertices[facets[j].edges[i].e - connect_offset].z;
                vtxn->normal.x = vertex_normals[facets[j].edges[i].e - connect_offset].x;
                vtxn->normal.y = vertex_normals[facets[j].edges[i].e - connect_offset].y;
                vtxn->normal.z = vertex_normals[facets[j].edges[i].e - connect_offset].z;
            }
        }
        else {
            PEXCoord *vtx;

            vtx = (PEXCoord *)malloc( vertex_data[j].count *
                                      sizeof(PEXCoord) );
            vertex_data[j].vertices.no_data = vtx;
            for(i = vertex_data[j].count-1; i >= 0; i--, vtx++) {
                vtx->x = vertices[facets[j].edges[i].e - connect_offset].x;
                vtx->y = vertices[facets[j].edges[i].e - connect_offset].y;
                vtx->z = vertices[facets[j].edges[i].e - connect_offset].z;
            }
        }

        PEXFillAreaWithData( display, object, PEXOCStore,
            shape_hint, ignore_edges,
            facet_attributes, vertex_attributes,
            color_type,
            &facet_data[j],
            vertex_data[j].count, vertex_data[j].vertices );
    }
    }
    else /* COUNTERCLOCKWISE */ {
    for(j = 0; j < nfacets; j++) {

        if( (facet_attributes & PEXGANormal) &&
            (facet_attributes & PEXGAColor) ) {
            facet_data[j].rgb_normal.normal.x = -facet_normals[j].x;
            facet_data[j].rgb_normal.normal.y = -facet_normals[j].y;
            facet_data[j].rgb_normal.normal.z = -facet_normals[j].z;
            facet_data[j].rgb_normal.rgb.red  = facet_colors[j].red;
            facet_data[j].rgb_normal.rgb.green= facet_colors[j].green;
            facet_data[j].rgb_normal.rgb.blue = facet_colors[j].blue;
        }
        else if( facet_attributes & PEXGANormal ) {
            facet_data[j].normal.x = -facet_normals[j].x;
            facet_data[j].normal.y = -facet_normals[j].y;
            facet_data[j].normal.z = -facet_normals[j].z;
        }
        else if( facet_attributes & PEXGAColor ) {
            facet_data[j].rgb.red  = facet_colors[j].red;
            facet_data[j].rgb.green= facet_colors[j].green;
            facet_data[j].rgb.blue = facet_colors[j].blue;
        }

        vertex_data[j].count = facets[j].count;

        if( (vertex_attributes & PEXGANormal) &&
            (vertex_attributes & PEXGAColor) ) {
            PEXVertexRGBNormal *vtxrn;

            vtxrn = (PEXVertexRGBNormal *) malloc( vertex_data[j].count *
                                               sizeof( PEXVertexRGBNormal ) ) ;
            vertex_data[j].vertices.rgb_normal = vtxrn;
            for(i = 0; i < vertex_data[j].count; i++, vtxrn++) {
               vtxrn->point.x  = vertices[facets[j].edges[i].e - connect_offset].x;
               vtxrn->point.y  = vertices[facets[j].edges[i].e - connect_offset].y;
               vtxrn->point.z  = vertices[facets[j].edges[i].e - connect_offset].z;
               vtxrn->rgb.red  = vertex_colors[facets[j].edges[i].e - connect_offset].red;
               vtxrn->rgb.green= vertex_colors[facets[j].edges[i].e - connect_offset].green;
               vtxrn->rgb.blue = vertex_colors[facets[j].edges[i].e - connect_offset].blue;
               vtxrn->normal.x = vertex_normals[facets[j].edges[i].e - connect_offset].x;
               vtxrn->normal.y = vertex_normals[facets[j].edges[i].e - connect_offset].y;
               vtxrn->normal.z = vertex_normals[facets[j].edges[i].e - connect_offset].z;
            }
        }
        else if( vertex_attributes & PEXGAColor ) {
            PEXVertexRGB *vtxr;

            vtxr = (PEXVertexRGB *) malloc( vertex_data[j].count *
                                               sizeof( PEXVertexRGB ) ) ;
            vertex_data[j].vertices.rgb = vtxr;
            for(i = 0; i < vertex_data[j].count; i++, vtxr++) {
                vtxr->point.x  = vertices[facets[j].edges[i].e - connect_offset].x;
                vtxr->point.y  = vertices[facets[j].edges[i].e - connect_offset].y;
                vtxr->point.z  = vertices[facets[j].edges[i].e - connect_offset].z;
                vtxr->rgb.red  = vertex_colors[facets[j].edges[i].e - connect_offset].red;
                vtxr->rgb.green= vertex_colors[facets[j].edges[i].e - connect_offset].green;
                vtxr->rgb.blue = vertex_colors[facets[j].edges[i].e - connect_offset].blue;
            }
        }
        else if( vertex_attributes & PEXGANormal ) {
            PEXVertexNormal *vtxn;

            vtxn = (PEXVertexNormal *) malloc( vertex_data[j].count *
                                               sizeof( PEXVertexNormal ) ) ;
            vertex_data[j].vertices.normal = vtxn;
            for(i = 0; i < vertex_data[j].count; i++, vtxn++) {
                vtxn->point.x  = vertices[facets[j].edges[i].e - connect_offset].x;
                vtxn->point.y  = vertices[facets[j].edges[i].e - connect_offset].y;
                vtxn->point.z  = vertices[facets[j].edges[i].e - connect_offset].z;
                vtxn->normal.x = vertex_normals[facets[j].edges[i].e - connect_offset].x;
                vtxn->normal.y = vertex_normals[facets[j].edges[i].e - connect_offset].y;
                vtxn->normal.z = vertex_normals[facets[j].edges[i].e - connect_offset].z;
            }
        }
        else {
            PEXCoord *vtx;

            vtx = (PEXCoord *)malloc( vertex_data[j].count *
                                      sizeof(PEXCoord) );
            vertex_data[j].vertices.no_data = vtx;
            for(i = 0; i < vertex_data[j].count; i++, vtx++) {
                vtx->x = vertices[facets[j].edges[i].e - connect_offset].x;
                vtx->y = vertices[facets[j].edges[i].e - connect_offset].y;
                vtx->z = vertices[facets[j].edges[i].e - connect_offset].z;
            }
        }
        PEXFillAreaWithData( display, object, PEXOCStore,
            shape_hint, ignore_edges,
            facet_attributes, vertex_attributes,
            color_type,
            &facet_data[j],
            vertex_data[j].count, vertex_data[j].vertices );
    }
    }

    free( facet_data );
    for(j = 0; j < nfacets; j++) {
             if( (vertex_attributes & PEXGANormal) &&
                 (vertex_attributes & PEXGAColor) )
             free( vertex_data[j].vertices.rgb_normal );
        else if( vertex_attributes & PEXGANormal )
             free( vertex_data[j].vertices.normal );
        else if( vertex_attributes & PEXGAColor )
             free( vertex_data[j].vertices.rgb );
        else
             free( vertex_data[j].vertices.no_data );
    }
    free( vertex_data );
}
